//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "services.h"

int loadService(const QString &desc, const QString &user, const QString &servName, const QString &workDir, const QString &recPath, bool start)
{
    Q_UNUSED(recPath);

    QFile file("/lib/systemd/system/" + servName + ".service");
    
    auto ret    = 0;
    auto exists = file.exists();

    if (!file.open(QFile::ReadWrite | QFile::Truncate))
    {
        QTextStream(stderr) << "err: failed to open service file: " << file.fileName() << " for writing. reason: " << file.errorString() << Qt::endl;

        ret = EACCES;
    }
    else
    {
        if (exists)
        {
            if (ret == 0) ret = QProcess::execute("systemctl", {"stop", servName});
            if (ret == 0) ret = QProcess::execute("systemctl", {"disable", servName});
        }
        
        if (ret == 0)
        {
            file.write("[Unit]\n");
            file.write("Description=" + desc.toUtf8() + "\n");
            file.write("After=network.target\n\n");
            file.write("[Service]\n");
            file.write("WorkingDirectory=" + workDir.toUtf8() + "\n");
            file.write("Type=simple\n");
            file.write("User=" + user.toUtf8() + "\n");
            file.write("Restart=always\n");

            if (servName.contains("vid_loop"))
            {
                file.write("RuntimeMaxSec=61\n");
            }

            file.write("ExecStart=/usr/bin/env " + servName.toUtf8() + "\n\n");
            file.write("[Install]\n");
            file.write("WantedBy=multi-user.target\n");
            file.close();

            if (ret == 0) ret = QProcess::execute("systemctl", {"daemon-reload"});

            if (start)
            {
                if (ret == 0) ret = QProcess::execute("systemctl", {"enable", servName});
                if (ret == 0) ret = QProcess::execute("systemctl", {"start", servName});
            }
        }
    }
        
    file.close();
    
    return ret;
}

int loadSh(const QString &name, const QString &exeCmd, const QString &buffDir, const QString &outDir, const QString &servUser, const QString &servGroup)
{
    QFile file("/usr/bin/" + name);
    
    auto ret = 0;
    
    if (!file.open(QFile::ReadWrite | QFile::Truncate))
    {
        QTextStream(stderr) << "err: failed to open shell script file: " << file.fileName() << " for writing. reason: " << file.errorString() << Qt::endl;

        ret = EACCES;
    }
    else
    {
        file.write("#!/bin/sh\n\n");
        file.write("cd " + buffDir.toUtf8() + "\n");
        file.write(exeCmd.toUtf8() + "\n");
        file.close();

        mkPath(buffDir);
        mkPath(outDir);
        mkPath(buffDir + "/live");
        mkPath(buffDir + "/img");
        
        QProcess::execute("chmod", {"-v", "+x", file.fileName()});
        QProcess::execute("chown", {servUser + ":" + servGroup, "-R", buffDir});
    }
    
    file.close();
    
    return ret;
}

QString camCmdFromConf(shared_t *conf, CmdExeType type)
{
    QString ret = "";
    
    if (type == MAIN_LOOP)
    {
        ret += QString(APP_TARGET) + " -c " + conf->conf;
    }
    else
    {
        ret += "ffmpeg -hide_banner -y -i '" + conf->recordUri + "' -strftime 1 -strftime_mkdir 1 ";
        
        if (conf->recordUri.contains("rtsp"))
        {
            ret += "-rtsp_transport udp ";
        }

        if (conf->vidCodec != "copy")
        {
            ret += "-vf fps=" + QString::number(conf->recFps) + ",scale=" + conf->recScale + " ";
        }

        ret += "-vcodec " + conf->vidCodec + " ";
        ret += "-acodec " + conf->audCodec + " ";
        ret += "-reset_timestamps 1 -sc_threshold 0 -g 2 -force_key_frames \"expr:gte(t, n_forced * 2)\" -t 60 -segment_time 2 -f segment ";
        ret += conf->buffPath + "/live/" + QString(STRFTIME_FMT) + conf->streamExt;
    }
    
    return ret;
}

int loadServiceByConf(const QString &confFile, bool start)
{
    auto ret = 0;
    
    shared_t conf;

    if (!rdConf(confFile, &conf))
    {
        ret = conf.retCode;
    }
    else
    {   
        auto desc = QString(APP_NAME) + " - " + conf.camName;
        
        if (ret == 0) ret = loadSh(conf.servMainLoop, camCmdFromConf(&conf, MAIN_LOOP), conf.buffPath, conf.recPath, conf.servUser, conf.servGroup);
        if (ret == 0) ret = loadSh(conf.servVidLoop, camCmdFromConf(&conf, VID_LOOP), conf.buffPath, conf.recPath, conf.servUser, conf.servGroup);
        
        if (ret == 0) ret = loadService(desc, conf.servUser, conf.servMainLoop, conf.buffPath, conf.recPath, start);
        if (ret == 0) ret = loadService(desc, conf.servUser, conf.servVidLoop,  conf.buffPath, conf.recPath, start);
    }
    
    return ret;
}

int rmService(const QString &servName)
{
    auto path = "/lib/systemd/system/" + servName + ".service";
    auto ret  = 0;
    
    if (QFile::exists(path))
    {
        if (ret == 0) ret = QProcess::execute("systemctl", {"stop", servName});
        if (ret == 0) ret = QProcess::execute("systemctl", {"disable", servName});

        QFile::remove(path);
        QFile::remove("/usr/bin/" + servName);
    }
    
    return ret;
}

int rmServiceByConf(const QString &confFile)
{
    shared_t conf;
    
    if (rdConf(confFile, &conf))
    {
        conf.retCode = rmService(conf.servMainLoop);
        conf.retCode = rmService(conf.servVidLoop);

        QDir(conf.buffPath).removeRecursively();
    }
    
    return conf.retCode;
}

void servStatByConf(const QString &confFile)
{
    shared_t conf;
    
    if (rdConf(confFile, &conf))
    {
        QTextStream cout(stdout);
        
        cout << "--" << conf.camName << Qt::endl;
        cout << "  " << conf.servMainLoop << ": ";

        if (QFile::exists("/lib/systemd/system/" + conf.servMainLoop + ".service"))
        {
            QProcess::execute("systemctl", {"is-enabled", conf.servMainLoop}); cout << Qt::endl;
        }
        else
        {
            cout << "Not Installed" << Qt::endl;
        }

        cout << "  " << conf.servVidLoop << ": ";

        if (QFile::exists("/lib/systemd/system/" + conf.servVidLoop + ".service"))
        {
            QProcess::execute("systemctl", {"is-enabled", conf.servVidLoop}); cout << Qt::endl;
        }
        else
        {
            cout << "Not Installed" << Qt::endl;
        }
    }
}

void servStatByDir(const QString &path)
{
    auto files = lsFilesInDir(path);

    for (auto &&conf : files)
    {
        servStatByConf(QDir::cleanPath(path) + "/" + conf);
    }
}

int loadServiceByDir(const QString &path, bool start)
{
    auto ret   = 0;
    auto files = lsFilesInDir(path);
    
    for (auto &&conf : files)
    {
        if (ret == 0 ) ret = loadServiceByConf(QDir::cleanPath(path) + "/" + conf, start);
    }

    return ret;
}

int rmServiceByDir(const QString &path)
{
    auto ret   = 0;
    auto files = lsFilesInDir(path);
    
    for (auto &&conf : files)
    {
        if (ret == 0 ) ret = rmServiceByConf(QDir::cleanPath(path) + "/" + conf);
    }

    return ret;
}
